%{

/*
   conf.l -- PIKT lexer for the PIKT.conf file

   $Header: /home/cvs/ucgsb/pikt/src/lib/conf.l,v 1.9 1999/11/29 05:09:13 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#include "common.h"
#include "conf.tab.h"

#define RETURN(token) conflval.sval = mystrdup(yytext, "conflex"); return token

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%option noyywrap

%%

("//"|"#")[^\n]*\n		{ lineno++; }

^uid				{ confparintp = &uid;
				  RETURN(_PARINT);
				}
^master				{ confparstrp = &master;
				  RETURN(_PARSTR);
				}
^domain				{ confparstrp = &domain;
				  RETURN(_PARSTR);
				}
^access				{ confparstrp = &accessmode;
				  RETURN(_PARSTR);
				}

^master_address			{ confparstrp = &master_address;
				  RETURN(_PARSTR);
				}
^authenticate_by_master		{ confparintp = &isauth_by_master;
				  RETURN(_PARBOL);
				}
^authenticate_by_master_address { confparintp = &isauth_by_master_address;
				  RETURN(_PARBOL);
				}

^all_services			{ confparsrv = ALL_SERVICES;
				  RETURN(_PARSRV);
				}
^kill_piktd_service		{ confparsrv = KILL_PIKTD_SERVICE;
				  RETURN(_PARSRV);
				}
^kill_piktc_svc_service		{ confparsrv = KILL_PIKTC_SVC_SERVICE;
				  RETURN(_PARSRV);
				}
^kill_service			{ confparsrv = KILL_SERVICE;
				  RETURN(_PARSRV);
				}
^restart_piktd_service		{ confparsrv = RESTART_PIKTD_SERVICE;
				  RETURN(_PARSRV);
				}
^restart_piktc_svc_service	{ confparsrv = RESTART_PIKTC_SVC_SERVICE;
				  RETURN(_PARSRV);
				}
^restart_service		{ confparsrv = RESTART_SERVICE;
				  RETURN(_PARSRV);
				}
^call_back			{ confparsrv = CALL_BACK_SERVICE;
				  RETURN(_PARSRV);
				}
^disable_alerts_service		{ confparsrv = DISABLE_ALERTS_SERVICE;
				  RETURN(_PARSRV);
				}
^enable_alerts_service		{ confparsrv = ENABLE_ALERTS_SERVICE;
				  RETURN(_PARSRV);
				}
^show_alerts_status_service	{ confparsrv = SHOW_ALERTS_STATUS_SERVICE;
				  RETURN(_PARSRV);
				}
^execute_commands_service	{ confparsrv = EXECUTE_COMMANDS_SERVICE;
				  RETURN(_PARSRV);
				}
^execute_(alerts|scripts)_service { confparsrv = EXECUTE_ALERTS_SERVICE;
				  RETURN(_PARSRV);
				}
^install_(alerts|scripts)_service { confparsrv = INSTALL_ALERTS_SERVICE;
				  RETURN(_PARSRV);
				}
^install_programs_service	{ confparsrv = INSTALL_PROGRAMS_SERVICE;
				  RETURN(_PARSRV);
				}
^install_files_service		{ confparsrv = INSTALL_FILES_SERVICE;
				  RETURN(_PARSRV);
				}
^install_objects_service	{ confparsrv = INSTALL_OBJECTS_SERVICE;
				  RETURN(_PARSRV);
				}
^install_service		{ confparsrv = INSTALL_SERVICE;
				  RETURN(_PARSRV);
				}
^fetch_(alerts|scripts)_service	{ confparsrv = FETCH_ALERTS_SERVICE;
				  RETURN(_PARSRV);
				}
^fetch_programs_service		{ confparsrv = FETCH_PROGRAMS_SERVICE;
				  RETURN(_PARSRV);
				}
^fetch_files_service		{ confparsrv = FETCH_FILES_SERVICE;
				  RETURN(_PARSRV);
				}
^fetch_objects_service		{ confparsrv = FETCH_OBJECTS_SERVICE;
				  RETURN(_PARSRV);
				}
^fetch_service			{ confparsrv = FETCH_SERVICE;
				  RETURN(_PARSRV);
				}
^delete_(alerts|scripts)_service { confparsrv = DELETE_ALERTS_SERVICE;
				  RETURN(_PARSRV);
				}
^delete_programs_service	{ confparsrv = DELETE_PROGRAMS_SERVICE;
				  RETURN(_PARSRV);
				}
^delete_files_service		{ confparsrv = DELETE_FILES_SERVICE;
				  RETURN(_PARSRV);
				}
^delete_objects_service		{ confparsrv = DELETE_OBJECTS_SERVICE;
				  RETURN(_PARSRV);
				}
^delete_service			{ confparsrv = DELETE_SERVICE;
				  RETURN(_PARSRV);
				}
^list_(alerts|scripts)_service	{ confparsrv = LIST_ALERTS_SERVICE;
				  RETURN(_PARSRV);
				}
^list_programs_service		{ confparsrv = LIST_PROGRAMS_SERVICE;
				  RETURN(_PARSRV);
				}
^list_files_service		{ confparsrv = LIST_FILES_SERVICE;
				  RETURN(_PARSRV);
				}
^list_objects_service		{ confparsrv = LIST_OBJECTS_SERVICE;
				  RETURN(_PARSRV);
				}
^list_service			{ confparsrv = LIST_SERVICE;
				  RETURN(_PARSRV);
				}

^homdir				{ confparstrp = &homdir;
				  RETURN(_PARSTR);
				}
^bindir				{ confparstrp = &bindir;
				  RETURN(_PARSTR);
				}
^etcdir				{ confparstrp = &etcdir;
				  RETURN(_PARSTR);
				}
^libdir				{ confparstrp = &libdir;
				  RETURN(_PARSTR);
				}
^vardir				{ confparstrp = &vardir;
				  RETURN(_PARSTR);
				}
^altdir				{ confparstrp = &altdir;
				  RETURN(_PARSTR);
				}
^prgdir				{ confparstrp = &prgdir;
				  RETURN(_PARSTR);
				}
^objdir				{ confparstrp = &objdir;
				  RETURN(_PARSTR);
				}
^cfgdir				{ confparstrp = &cfgdir;
				  RETURN(_PARSTR);
				}
^hstdir				{ confparstrp = &hstdir;
				  RETURN(_PARSTR);
				}
^logdir				{ confparstrp = &logdir;
				  RETURN(_PARSTR);
				}

^rpc_timeout			{ confparintp = &rpc_timeout;
				  RETURN(_PARINT);
				}
^rpc_retry_timeout		{ confparintp = &rpc_retry_timeout;
				  RETURN(_PARINT);
				}
^rpcinfo_retries		{ confparintp = &rpcinfo_retries;
				  RETURN(_PARINT);
				}
^execcmdout_limit		{ confparintp = &execcmdout_limit;
				  RETURN(_PARINT);
				}
^verbose_log			{ confparintp = &isverboselog;
				  RETURN(_PARBOL);
				}
^auto_syslog			{ confparintp = &isautosyslog;
				  RETURN(_PARBOL);
				}
^tcp_only			{ confparintp = &istcponly;
				  RETURN(_PARBOL);
				}

^syslog_facility		{ return _SYSLOGFAC; }

yes|YES|true|TRUE|on|ON		{ conflval.ival = TRUE;
				  return _BOL;
				}
no|NO|false|FALSE|off|OFF	{ conflval.ival = FALSE;
				  return _BOL;
				}

-?{D}+				{ conflval.ival = atoi(yytext);
				  return _INT;
				}

[^ \t\n]+			{ RETURN(_STR); }

{W}				{ ; }

\n				{ lineno++; }

%%

/* conf error handler
 */
void
conferror (const char *message)
{

	fprintf(stderr, "PIKT.conf, at or near line %d: %s (in confparse()): %s\n", lineno, message, yytext);
	fflush(stderr);
	bailout = TRUE;
	return;
}
