%{

/*
   cmd.y -- PIKT piktc command-line parser

   $Header: /home/cvs/ucgsb/pikt/src/piktc/cmd.y,v 1.10 1999/10/31 05:20:21 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#include "../lib/common.h"
#include "piktc.h"

#define SETBS(a,b)	tbs1 = a; \
			a = b; \
			free(tbs1)

unsigned long *tbs1, *tbs2, *tbs3;

%}

%union {
	signed char *sval;
}

%token <sval> _DEFINE _ALERT _PROGRAM _FILE _OBJECT _HOST _COMMAND
%token _MINUS _PLUS _OPT _DEFOPT _ALTOPT _PRGOPT _FILOPT _OBJOPT _HSTOPT _ALL _CMDOPT

%%

start:		/* empty */ {
			if (!postscan) {
				ishelp = TRUE;	/* never reached in postscan */
			}
		}
		| options
		| options items
		;

options:	option
		| options option
		;

option:		_MINUS opts
		/* | _PLUS opts	*/	/* no plus opts now, maybe in future */
		;

opts:		_OPT
		| opts _OPT
		;

items:		item
		| items item
		;

item:		_PLUS initems
		| _MINUS exitems
		| _PLUS _CMDOPT commands
		;

initems:	_DEFOPT indefines
		| _DEFOPT {
			/* no defines */
			if (!postscan)
				bailout = TRUE;	/* never reached in postscan */
		}
		| _ALTOPT _ALL {
			if (!postscan)
				regname(&incmdalerts, "all", FALSE, FALSE);
		}
		| _ALTOPT inalerts
		| _ALTOPT {
			/* no alerts */
			if (!postscan)
				bailout = TRUE; /* never reached in postscan */
		}
		| _PRGOPT _ALL {
			if (!postscan )
				regname(&incmdprograms, "all", FALSE, FALSE);
		}
		| _PRGOPT inprograms
		| _PRGOPT {
			/* no programs */
			if (!postscan)
				bailout = TRUE;	/* never reached in postscan */
		}
		| _FILOPT _ALL {
			if (!postscan )
				regname(&incmdfiles, "all", FALSE, FALSE);
		}
		| _FILOPT infiles
		| _FILOPT {
			/* no files */
			if (!postscan)
				bailout = TRUE;	/* never reached in postscan */
		}
		| _OBJOPT _ALL {
			if (!postscan)
				regname(&incmdobjects, "all", FALSE, FALSE);
		}
		| _OBJOPT inobjects
		| _OBJOPT {
			/* no objects */
			if (!postscan)
				bailout = TRUE;	/* never reached in postscan */
		}
		| _HSTOPT _ALL {
			if (!postscan)
				regname(&incmdhosts, "all", FALSE, FALSE);
		}
		| _HSTOPT inhosts
		| _HSTOPT {
			/* no hosts */
			if (!postscan)
				bailout = TRUE;	/* never reached in postscan */
		}
		;

indefines:	indefine
		| indefines indefine
		;

indefine:	_DEFINE {
			if (postscan) {
				tbs3 = getbs($1, DEF);
				SETBS(gdbs, bsunion(gdbs, tbs3, dbslongs));
				tbs2 = bscomplement(tbs3, dbslongs);
				SETBS(gubs, bsintersection(gubs, tbs2, dbslongs));
				free(tbs2);
			}
			free($1);
		}
		;

inalerts:	_ALERT {
			if (!postscan)
				regname(&incmdalerts, $1, FALSE, FALSE);
			free($1);
		}
		| inalerts _ALERT {
			if (!postscan)
				regname(&incmdalerts, $2, FALSE, FALSE);
			free($2);
		}
		;

inprograms:	_PROGRAM {
			if (!postscan)
				regname(&incmdprograms, $1, FALSE, FALSE);
			free($1);
		}
		| inprograms _PROGRAM {
			if (!postscan)
				regname(&incmdprograms, $2, FALSE, FALSE);
			free($2);
		}
		;

infiles:	_FILE {
			if (!postscan)
				regname(&incmdfiles, $1, FALSE, FALSE);
			free($1);
		}
		| infiles _FILE {
			if (!postscan)
				regname(&incmdfiles, $2, FALSE, FALSE);
			free($2);
		}
		;

inobjects:	_OBJECT {
			if (!postscan)
				regname(&incmdobjects, $1, FALSE, FALSE);
			free($1);
		}
		| inobjects _OBJECT {
			if (!postscan)
				regname(&incmdobjects, $2, FALSE, FALSE);
			free($2);
		}
		;

inhosts:	_HOST {
			if (!postscan)
				regname(&incmdhosts, $1, FALSE, FALSE);
			free($1);
		}
		| inhosts _HOST {
			if (!postscan)
				regname(&incmdhosts, $2, FALSE, FALSE);
			free($2);
		}
		;

exitems:	_DEFOPT exdefines
		| _DEFOPT {
			/* no defines */
			if (!postscan)
				bailout = TRUE;	/* never reached in postscan */
		}
		| _ALTOPT _ALL {
			if (!postscan)
				regname(&excmdalerts, "all", FALSE, FALSE);
		}
		| _ALTOPT exalerts
		| _ALTOPT {
			/* no alerts */
			if (!postscan)
				bailout = TRUE;	/* never reached in postscan */
		}
		| _PRGOPT _ALL {
			if (!postscan)
				regname(&excmdprograms, "all", FALSE, FALSE);
		}
		| _PRGOPT exprograms
		| _PRGOPT {
			/* no programs */
			if (!postscan)
				bailout = TRUE;	/* never reached in postscan */
		}
		| _FILOPT _ALL {
			if (!postscan)
				regname(&excmdfiles, "all", FALSE, FALSE);
		}
		| _FILOPT exfiles
		| _FILOPT {
			/* no files */
			if (!postscan)
				bailout = TRUE;	/* never reached in postscan */
		}
		| _OBJOPT _ALL {
			if (!postscan)
				regname(&excmdobjects, "all", FALSE, FALSE);
		}
		| _OBJOPT exobjects
		| _OBJOPT {
			/* no objects */
			if (!postscan)
				bailout = TRUE;	/* never reached in postscan */
		}
		| _HSTOPT _ALL {
			if (!postscan)
				regname(&excmdhosts, "all", FALSE, FALSE);
		}
		| _HSTOPT exhosts
		| _HSTOPT {
			/* no hosts */
			if (!postscan)
				bailout = TRUE;	/* never reached in postscan */
		}
		;

exdefines:	exdefine
		| exdefines exdefine
		;

exdefine:	_DEFINE {
			if (postscan) {
				tbs3 = getbs($1, DEF);
				SETBS(gubs, bsunion(gubs, tbs3, dbslongs));
				tbs2 = bscomplement(tbs3, dbslongs);
				SETBS(gdbs, bsintersection(gdbs, tbs2, dbslongs));
				free(tbs2);
			}
			free($1);
		}
		;

exalerts:	_ALERT {
			if (!postscan)
				regname(&excmdalerts, $1, FALSE, FALSE);
			free($1);
		}
		| exalerts _ALERT {
			if (!postscan)
				regname(&excmdalerts, $2, FALSE, FALSE);
			free($2);
		}
		;

exprograms:	_PROGRAM {
			if (!postscan)
				regname(&excmdprograms, $1, FALSE, FALSE);
			free($1);
		}
		| exprograms _PROGRAM {
			if (!postscan)
				regname(&excmdprograms, $2, FALSE, FALSE);
			free($2);
		}
		;

exfiles:	_FILE {
			if (!postscan)
				regname(&excmdfiles, $1, FALSE, FALSE);
			free($1);
		}
		| exfiles _FILE {
			if (!postscan)
				regname(&excmdfiles, $2, FALSE, FALSE);
			free($2);
		}
		;

exobjects:	_OBJECT {
			if (!postscan)
				regname(&excmdobjects, $1, FALSE, FALSE);
			free($1);
		}
		| exobjects _OBJECT {
			if (!postscan)
				regname(&excmdobjects, $2, FALSE, FALSE);
			free($2);
		}
		;

exhosts:	_HOST {
			if (!postscan)
				regname(&excmdhosts, $1, FALSE, FALSE);
			free($1);
		}
		| exhosts _HOST {
			if (!postscan)
				regname(&excmdhosts, $2, FALSE, FALSE);
			free($2);
		}
		;

commands:	command
		| commands command
		;

command:	_COMMAND {
			if (!postscan)
				regname(&incmdcommands, $1, TRUE, FALSE);
			free($1);
		}
		;

%%
