%{

/*
   pikts.l -- PIKT filter utility for stripping away comments

   $Header: /home/cvs/ucgsb/pikt/src/piktc/pikts.l,v 1.10 1999/10/17 05:59:39 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA

   --------------------------------------------------------------------------

   Usage:  pikts [-bGhlT]

           -b              strip away block comments only
           -G              run in debug mode
           -h              show program help
           -l              strip away line comments only
           -T              run in test mode

           note:  the default is to strip away both block and line comments

 */

#include "../lib/common.h"
#include "piktc.h"

const char pikts_version[] = "PIKT pikts version " VERSION;

int isdebug = FALSE;
int ishelp = FALSE;
int istest = FALSE;
int isblockstrip = FALSE;
int islinestrip = FALSE;

int bailout = FALSE;	/* signifies command-line error */

int piktslex(void);
void pikts_usage(void);
int isopt(char **, char **, int);

%}

W	[ \t]

%x CMNT0 CMNT1

%option noyywrap

%%

<INITIAL>("\\/*"|"/\\*"|"\\/\\*") { printf("/*"); }
<INITIAL>("\\*/"|"*\\/"|"\\*\\/") { printf("*/"); }
<INITIAL>("\\//"|"/\\/"|"\\/\\/") { printf("//"); }

<CMNT0,CMNT1>("\\/*"|"/\\*"|"\\/\\*") { ; }
<CMNT0,CMNT1>("\\*/"|"*\\/"|"\\*\\/") { ; }
<CMNT0,CMNT1>("\\//"|"/\\/"|"\\/\\/") { ; }

"/*"				{ if (isblockstrip)
					BEGIN(CMNT0);
				  else
					ECHO;
				}
{W}*"//"			{ if (islinestrip)
					BEGIN(CMNT1);
				  else
					ECHO;
				}
\n				{ ECHO; }
.				{ ECHO; }

<CMNT0>"*/"			{ BEGIN(INITIAL); }
<CMNT0>\n			{ ECHO; }
<CMNT0>.			{ ; }

<CMNT1>.+			{ ; }
<CMNT1>\n			{ BEGIN(INITIAL);
				  ECHO;
				}

%%

int
main (int argc, char *argv[])
{
	istest = isopt(argv+1, argv+argc-1, 'T');
	isdebug = isopt(argv+1, argv+argc-1, 'G');
	if((ishelp = isopt(argv+1, argv+argc-1, 'h'))) {
		pikts_usage();
		exit(1);
	}
	isblockstrip = isopt(argv+1, argv+argc-1, 'b');
	islinestrip = isopt(argv+1, argv+argc-1, 'l');
	if (!isblockstrip && !islinestrip) {	/* with no command-line args,
						   default to strip away all
						   comments
						 */
		isblockstrip = islinestrip = TRUE;
	}
	piktslex();
	exit(0);
}

/* output program usage to screen
 */
void
pikts_usage (void)
{
	printf("\nUsage:  pikts [-bGhlT]\n\n");
	printf("\t-b\t\t\tstrip away block comments only\n");
	printf("\t-G\t\t\trun in debug mode\n");
	printf("\t-h\t\t\tshow program help\n");
	printf("\t-l\t\t\tstrip away line comments only\n");
	printf("\t-T\t\t\trun in test mode\n");
	printf("\n\tnote:  the default is to strip away both block and line comments\n\n");
	return;
}

/* look for a command-line (minus) switch, returning TRUE if found, else
   FALSE; this is for simpler command-line parsing; for more involved parsing,
   we use cmd.[lyc] and cmdparse()
 */
int
isopt (char **avcurr, char **avlast, int opt)
{
	while (avcurr <= avlast) {
		if (**avcurr == '-') {
			if (strchr(*avcurr, opt))
				return TRUE;
		}
		avcurr++;
	}
	return FALSE;
}
