%{

/*
   sys.l -- PIKT piktc lexer for the systems.cfg file

   $Header: /home/cvs/ucgsb/pikt/src/piktc/sys.l,v 1.13 1999/11/14 18:34:08 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#include "../lib/common.h"
#include "piktc.h"
#include "sys.tab.h"

char _syslex_[] = "syslex";

#define RETURN(token) syslval.sval = mystrdup(yytext, _syslex_); return token

char *cp;

typedef struct {
	char *file;		/* we should maybe garbage collect these
				   after preprocess()'ing, but this won't
				   gain us much, so why bother
				 */
	FILE *yyin;
	int fd[2];
	YY_BUFFER_STATE bs;
	pid_t pid;
	int lineno;
} incfil;

incfil sysincfiles[MAXIFS+1];

char *sysincfile = (char *) NULL;

void beginsysincfile(char *);
void endsysincfile(void);
void pushsysincfile(char *, FILE *, int *, YY_BUFFER_STATE, pid_t, int);
void popsysincfile(char **, FILE **, int *, YY_BUFFER_STATE *, pid_t *, int *);

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%%

^{L}[-_\.0-9a-zA-Z]*		{ RETURN(_STANZA); }

^#{W}*include{W}+\<[^\>]+\>{W}*(\177)?\n? { lineno++;
				  *strrchr(yytext, '>') = '\0';
				  cp = strchr(yytext, '<') + 1;
				  /* linux chokes on this
				  FREE(sysincfile);
				   */
				  sysincfile = (char *) mymalloc((strlen(cfgdir)+strlen(cp)+2) * sizeof(char), _syslex_);
				  sprintf(sysincfile, "%s/%s", cfgdir, cp);
				  beginsysincfile(sysincfile);
				}

^#{W}*include{W}+\"[^\"]+\"{W}*(\177)?\n? { lineno++;
				  *strrchr(yytext, '\"') = '\0';
				  cp = strchr(yytext, '\"') + 1;
				  /* linux chokes on this
				  FREE(sysincfile);
				   */
				  sysincfile = mystrdup(cp, _syslex_);
				  beginsysincfile(sysincfile);
				}

^#.*(\177)?\n			{ lineno++;
				  syserror("syntax error");
				}

^{W}+hosts			{ return _HOSTS; }

^{W}+members			{ return _MEMBERS; }

^{W}+aliases			{ return _ALIASES; }

{L}[-_\.0-9a-zA-Z]*		{ RETURN(_HOST); }

{W}				{ ; }

(\177)?\n?			{ lineno++; }

.				{ return yytext[0]; }

%%

/* sys error handler
 */
void
syserror (const char *message)
{

	if (stzerr)
		return;
	else {
		stzerr = TRUE;
		fprintf(stderr, "%s, at or near line %d: %s (in sysparse()): %s\n", cfgfilenames[cfgcurr], lineno, message, yytext ? yytext : "");
		fflush(stderr);
		bailout = TRUE;
		return;
	}
}


/* sys end-of-file handler
 */
int
syswrap (void)
{

	if (incnum > 0) {
		endsysincfile();
		return 0;
	} else
		return 1;
}

/* push the current input buffer onto the file stack; create and switch to
   a new input buffer; open the new include file and start feeding its
   contents to the new input buffer; if stripping comments from the new
   include file, spawn a child process to pass the include file
   contents to the strip lexer
 */
void
beginsysincfile (char *file)
{
	struct stat fbuf;

	if (stat(file, &fbuf) != 0) {
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in beginsysincfile(), stat() error for %s", file);
	}
	pushsysincfile(infile, sysin, fd, YY_CURRENT_BUFFER, pid, lineno);
	/* linux possibly chokes on this
	FREE(cfgfileprev);
	 */
	cfgfileprev = cfgfilecurr;
	cfgfilecurr = infile = mystrdup(file, "beginsysincfile");
	incnum++;
	if (incnum > MAXIFS)
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in beginsysincfile(), maximum number (%d) of include files exceeded", MAXIFS);
	lineno = 1;
	if (dostrip) {
		if (pipe(fd) < 0)
			logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in beginsysincfile(), pipe() error");
		/* open the ipc channels */
		ipcopen(incnum);
		switch (pid = fork()) {
			case -1:	/* error */
				logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in beginsysincfile(), fork() error");
				break;
			/* pipe from child to parent */
			case 0:		/* child */
				close(fd[0]);
				stripin = fopen(infile, "r");
				stripout = fdopen(fd[1], "w");
				wait_parent(incnum);
				tell_parent(incnum);
				striplex();
				fclose(stripin);
				fclose(stripout);
				close(fd[1]);
				tell_parent(incnum);
				wait_parent(incnum);
				_exit(0);
				break;
			default:	/* parent */
				close(fd[1]);
				sysin = fdopen(fd[0], "r");
				tell_child(incnum);
				wait_child(incnum);
				break;
		}
	} else {
		pid = 0;
		fd[0] = -1;
		fd[1] = -1;
		sysin = fopen(infile, "r");
	}
	yy_switch_to_buffer(yy_create_buffer(sysin, MAXBUF));
	return;
}

/* at the end of the current include file, close the file, delete its
   input buffer, pop the previous include file off the file stack, and
   switch to the previous input buffer
 */
void
endsysincfile (void)
{
	YY_BUFFER_STATE bs;
	int status;

	fclose(sysin);
	if (dostrip) {
		close(fd[0]);
		wait_child(incnum);
		tell_child(incnum);
		waitpid(pid, &status, 0);
		/* close the ipc channels */
		ipcclose(incnum);
	}
	incnum--;
	if (incnum < 0)
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in endsysincfile(), incnum < 0");
	yy_delete_buffer(YY_CURRENT_BUFFER);
	popsysincfile(&infile, &sysin, fd, &bs, &pid, &lineno);
	/* linux possibly chokes on this
	FREE(cfgfileprev);
	 */
	cfgfileprev = cfgfilecurr;
	cfgfilecurr = infile;
	yy_switch_to_buffer(bs);
	return;
}

/* push current input buffer info onto the file stack
 */
void
pushsysincfile (char *file, FILE *yyin, int *fd, YY_BUFFER_STATE bs, pid_t pid, int lno)
{

	sysincfiles[incnum].file = mystrdup(file, "pushsysincfile");
	sysincfiles[incnum].yyin = yyin;
	sysincfiles[incnum].fd[0] = *fd;
	sysincfiles[incnum].fd[1] = *(fd+1);
	sysincfiles[incnum].bs = bs;
	sysincfiles[incnum].pid = pid;
	sysincfiles[incnum].lineno = lno;
	return;
}

/* pop input buffer info off the file stack
 */
void
popsysincfile (char **file, FILE **yyin, int *fdp, YY_BUFFER_STATE *bsp, pid_t *pidp, int *lnp)
{

	*file = sysincfiles[incnum].file;
	*yyin = sysincfiles[incnum].yyin;
	*fdp = sysincfiles[incnum].fd[0];
	*(fdp+1) = sysincfiles[incnum].fd[0];
	*bsp = sysincfiles[incnum].bs;
	*pidp = sysincfiles[incnum].pid;
	*lnp = sysincfiles[incnum].lineno;
	return;
}
