/*
   alt.c -- PIKT alarm script processing routines

   $Header: /home/cvs/ucgsb/pikt/src/piktd/alt.c,v 1.15 1999/11/05 03:21:08 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include "../lib/common.h"
#include "piktd.h"

signed int vrndxhi;

void initvals(val **, char *);
void initvar(int, int);
int freevarslot(void);

/* register an alarm: initialize its data structures; add it to the alarms
   list
 */
void
regalm (char *name)
{
	alrm *nalp;
	register int i;

	nalp = (alrm *) mymalloc(sizeof(alrm), "regalm");
	nalp->next = (alrm *) NULL;
	nalp->name = name;
	nalp->status = INACTIVE;
	nalp->level = 0;
	nalp->task = (char *) NULL;
	nalp->input.type = 0;
	nalp->input.source = (char *) NULL;
	nalp->input.logname = (char *) NULL;
	nalp->input.logino = 0L;
	nalp->input.logsize = 0L;
	nalp->input.logposend = 0L;
	nalp->separators = (char *) NULL;
	nalp->filter = (char *) NULL;
	nalp->regexp = (char *) NULL;
	nalp->datcnt = 0;
	nalp->maxdatndx = 0;
	nalp->postype = 0;
	for (i=0; i<=MAXVARS; i++) {
		nalp->varname[i] = (char *) NULL;
		nalp->valtype[i] = NIL;
		nalp->vartype[i] = NIL;
		nalp->dattype[i] = NIL;
		nalp->maxarndx[i] = 0;
		nalp->oldhst[i]  = FALSE;
		nalp->newhst[i]  = FALSE;
		nalp->vars[i].s = (spcscl *) NULL;
		nalp->vars[i].a = (spcary *) NULL;
		nalp->vars[i].f = (FILE *) NULL;
	}
	initvals(&nalp->values, name);
	initvals(&nalp->values->next, name);
	initvals(&nalp->values->next->next, name);
	vrndxhi = 0;
	nalp->scvars = (ky *) NULL;
	nalp->keys = (ky *) NULL;
	nalp->begin = (rul *) NULL;
	nalp->rules = (rul *) NULL;
	nalp->end = (rul *) NULL;
	nalp->hsttyp = 0;
	nalp->hsterr = TRUE;		/* guilty until proven innocent */
	if (!alarms)
		alarms = nalp;
	else
		alarmcurr->next = nalp;
	alarmcurr = nalp;
	return;
}

/* create and initialize the vals data structure
 */
void
initvals(val **values, char *name)
{
	val *vlp;
	register int i;

	vlp = (val *) mymalloc(sizeof(val), "initvals");
	vlp->prev = (val *) NULL;
	vlp->next = (val *) NULL;
	vlp->key = (char *) NULL;
	for (i=0; i<=MAXVARS; i++) {
		vlp->vals[i].s = (chrscl *) NULL;
		vlp->vals[i].a = (chrary *) NULL;
	}
	*values = vlp;
	return;
}

/* register a variable: if in a dat statement, process its positional info;
   initialize its data structures; set its valtype, vartype & dattype
 */
void
regvar (char *name, int valtype, int vartype, int dattype, int arndx, char *position)
{
	char *cp;
	int i, n;
	signed int vrndx;
	char pos[16+1];
	int pl;
	int trimtyp;

	trimtyp = NEITHER;
	vrndx = varndx(name);
	if (position) {		/* hence, a var in a dat statement */
		/* check if listed in previous dat statement */
		switch (vrndx) {
			case NIL:
			case INLINNOW:
				break;
			/* varndx() never returns INLINPRC, and regvar() is
			   never called with a varprc
			case INLINPRC:
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, attempted registration of @inlin", ALMNAM);
				STATUS = INACTIVE;
				return;
				break;
			 */
			case VLPRS:
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, $%s is a parse var", ALMNAM, name);
				STATUS = INACTIVE;
				return;
				break;
			case ERROR:
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, attempted use of reserved fn name %s as var", ALMNAM, name);
				STATUS = INACTIVE;
				return;
				break;
			case RDLINNOW:
			/* varndx() never returns RDLINPRC
			case RDLINPRC:
			 */
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, %s is a reserved var name", ALMNAM, name);
				STATUS = INACTIVE;
				return;
				break;
			default:
				switch (vartype) {
					case SCALAR:
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, var %c%s already specified in previous dat statement", ALMNAM, dattype ? '#' : '$', name);
						STATUS = INACTIVE;
						return;
						break;
					case ARRAY:
						if (alarmcurr->vars[vrndx].a && (TRIM(vrndx,arndx) != NIL)) {
							logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, var %c%s[%d] already specified in previous dat statement", ALMNAM, dattype ? '#' : '$', name, arndx);
							STATUS = INACTIVE;
							return;
						}
						break;
					case NIL:
						break;
					default:
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, bad vartype (%d) for var %s", ALMNAM, vartype, name);
						STATUS = INACTIVE;
						return;
						break;
				}
				break;
		}
		if (*position == '[') {
			pl = strlen(position);
			if (*(position+pl-1) == ']') {
				*(position+pl-1) = '\0';
				strcpy(pos,position+1);
				trimtyp = BOTH;
			} else
				POSTYP = ERROR;
		} else if (strchr(position,']')) {
				POSTYP = ERROR;
		} else {
			strcpy(pos,position);
			trimtyp = NEITHER;
		}
		if (POSTYP == ERROR) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, improperly formed data position:  %s", ALMNAM, position);
			STATUS = INACTIVE;
			return;
		}
		if ((cp = strchr(pos,','))) {
			if (!POSTYP)
				POSTYP = COLUMNS;
			else if (POSTYP != COLUMNS) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, inconsistent data position: %s", ALMNAM, pos);
				STATUS = INACTIVE;
				return;
			}
		} else if (isdigit((int)pos[0]) || (pos[0] == '$')) {
			if (!POSTYP)
				POSTYP = FIELDS;
			else if (POSTYP != FIELDS) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, inconsistent data position: %s", ALMNAM, pos);
				STATUS = INACTIVE;
				return;
			}
		} else {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, bad data position: %s", ALMNAM, pos);
			STATUS = INACTIVE;
			return;
		}
		/* at this point, have unique var name, valid pos */
		DATCNT++;
		if ((vrndx == NIL) || (vrndx == INLINNOW))
			vrndx = ++vrndxhi;
		/* check if max var count exceeded */
		if (vrndxhi > MAXVARS) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, maximum var count (%d) exceeded for var %c%s%s", ALMNAM, MAXVARS, dattype ? '#' : '$', name, vartype ? "[]" : "");
			STATUS = INACTIVE;
			return;
		}
		/* check to see if data slot already taken */
		if (VARNAM(vrndx) && (vartype == SCALAR)) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, data slot %d already taken", ALMNAM, vrndx);
			STATUS = INACTIVE;
			return;
		}
		initvar(vartype, vrndx);
		SETSPC(trim,vrndx,arndx,trimtyp);
		if (POSTYP == COLUMNS) {
			*cp = '\0';
			if ((i = atoi(pos)) && (i > 0) && (i <= MAXLINE)) {
				SETSPC(begincol,vrndx,arndx,i);
				if (!strcmp(cp+1,"$")) {
					SETSPC(endcol,vrndx,arndx,MAXLINE);
				} else if ((i = atoi(cp+1)) && (i > 0) && (i <= MAXLINE)) {
					if (vartype == SCALAR) {
						if (i >= BEGINCOL(vrndx,0))
							alarmcurr->vars[vrndx].s->endcol[0] = i;
					} else {
						if (i >= BEGINCOL(vrndx,arndx))
							alarmcurr->vars[vrndx].a->subary[SA(arndx)]->endcol[SN(arndx)] = i;
					}
				} else {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, bad data position: %s", ALMNAM, pos);
					STATUS = INACTIVE;
					return;
				}
			} else {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, bad data position: %s", ALMNAM, pos);
				STATUS = INACTIVE;
				return;
			}
			regscvar(vrndx, arndx);
		} else if (pos[0] == '$') {
			if (pos[1]) {
				if (pos[1] == '-') {
					errno = 0;
					n = atoi(pos+2);
					if (errno) {
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, bad data position: %s", ALMNAM, pos);
						STATUS = INACTIVE;
						return;
					}
					SETSPC(fldnum,vrndx,arndx,-(n+1));
				} else {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, bad data position: $s", ALMNAM, pos);
					STATUS = INACTIVE;
					return;
				}
			} else
				SETSPC(fldnum,vrndx,arndx,-1);
			if (vartype == SCALAR)
				MAXDATNDX = maxval(MAXDATNDX, -FLDNUM(vrndx,0));
			else
				MAXDATNDX = maxval(MAXDATNDX, -FLDNUM(vrndx,arndx));
		} else if (isdigit((int)pos[0])) {
			errno = 0;
			n = atoi(pos);
			if (errno) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, bad data position: %s", ALMNAM, pos);
				STATUS = INACTIVE;
				return;
			}
			SETSPC(fldnum,vrndx,arndx,n);
			if (vartype == SCALAR)
				MAXDATNDX = maxval(MAXDATNDX, FLDNUM(vrndx,0));
			else
				MAXDATNDX = maxval(MAXDATNDX, FLDNUM(vrndx,arndx));
		}
	/* else in the remaining cases, no position, hence a var not in
	   a dat statement, e.g., in a set or for statement, or a
	   file handle
	 */
	} else {
		vrndx = varndx(name);
		if (stmcurr)
			stmcurr->vrndx = vrndx;
		if ((vrndx == NIL) || (DATTYP(vrndx) == NIL)) {
			if (vrndx == NIL) {
				vrndxhi = vrndx = freevarslot();
				if (stmcurr)
					stmcurr->vrndx = vrndx;
				if (vrndx == NIL) {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, maximum var count (%d) exceeded for var %c%s%s", ALMNAM, MAXVARS, dattype ? '#' : '$', name, vartype ? "[]" : "");
					STATUS = INACTIVE;
					return;
				}
			}
			if (dattype == FILEHANDLE) {
				FLHNDL(vrndx) = (FILE *) NULL;
			} else {
				initvar(vartype, vrndx);
			}
		} else if (vrndx == ERROR) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, attempted use of reserved fn name %s as var", ALMNAM, name);
			STATUS = INACTIVE;
			return;
		} else if ((vrndx == RDLINNOW) || (vrndx == INLINNOW)) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, %s is a reserved var name", ALMNAM, name);
			STATUS = INACTIVE;
			return;
		} else { /* valid var name seen before */
			MAXARNDX(vrndx) = maxval(MAXARNDX(vrndx), arndx);
			VALTYP(vrndx) |= valtype;
			return;
		}
	}
	if (vrndx <= 0)	/* NIL, INLINNOW */
		vrndx = vrndxhi;
	if (VARNAM(vrndx)) {
		if (strcmp(VARNAM(vrndx), name)) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, inconsistent var names for vrndxhi %d: %s %s", ALMNAM, vrndx, VARNAM(vrndx), name);
			STATUS = INACTIVE;
			return;
		}
	} else
		VARNAM(vrndx) = mystrdup(name, "regvar");
	if (VALTYP(vrndx) == NIL)
		VALTYP(vrndx) = valtype;
	else if ((VALTYP(vrndx) & FIXEDVAL) && (VALTYP(vrndx) != valtype)) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, inconsistent val type for %s%s", ALMNAM, name, vartype ? "[]" : "");
		STATUS = INACTIVE;
		return;
	} else	/* should never be reached */
		VALTYP(vrndx) |= valtype;
	if (VARTYP(vrndx) == NIL)
		VARTYP(vrndx) = vartype;
	/* saw, e.g., $f before, $f[] now, or vice-versa */
	else if (VARTYP(vrndx) != vartype) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, inconsistent var type for %s%s", ALMNAM, name, vartype ? "[]" : "");
		STATUS = INACTIVE;
		return;
	}
	if (DATTYP(vrndx) == NIL)
		DATTYP(vrndx) = dattype;
	/* saw, e.g., $f[] before, #f[] now, or vice-versa */
	else if (DATTYP(vrndx) != dattype) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regvar(), %s, inconsistent dat type for %s%s", ALMNAM, name, vartype ? "[]" : "");
		STATUS = INACTIVE;
		return;
	}
	MAXARNDX(vrndx) = maxval(MAXARNDX(vrndx), arndx);
	return;
}

/* create and initialize a variable's data structures
 */
void
initvar (int vartype, int vrndx)
{
	static char _initvar_[] = "initvar";
	spcscl *ssp;
	spcary *sap;
	chrscl *csp;
	chrary *cap;
	register int i;

	switch (vartype) {
		case SCALAR:
			csp = alarmcurr->values->vals[vrndx].s = (chrscl *) mymalloc(sizeof(chrscl), _initvar_);
			csp->val[0] = (char *) NULL;
			csp = alarmcurr->values->next->vals[vrndx].s = (chrscl *) mymalloc(sizeof(chrscl), _initvar_);
			csp->val[0] = (char *) NULL;
			csp = alarmcurr->values->next->next->vals[vrndx].s = (chrscl *) mymalloc(sizeof(chrscl), _initvar_);		/* VALGLB */
			csp->val[0] = (char *) NULL;
			ssp = alarmcurr->vars[vrndx].s = (spcscl *) mymalloc(sizeof(spcscl), _initvar_);
			ssp->fldnum[0] = 0;
			ssp->begincol[0] = 0;
			ssp->endcol[0] = 0;
			ssp->trim[0] = NIL;
			break;
		case ARRAY:
			if (!alarmcurr->values->vals[vrndx].a) {
				cap = alarmcurr->values->vals[vrndx].a = (chrary *) mymalloc(sizeof(chrary), _initvar_);
				for (i=0; i<MAXSUBARY; i++)
					cap->subary[i] = (chrsubary *) NULL;
				cap = alarmcurr->values->next->vals[vrndx].a = (chrary *) mymalloc(sizeof(chrary), _initvar_);
				for (i=0; i<MAXSUBARY; i++)
					cap->subary[i] = (chrsubary *) NULL;
				cap = alarmcurr->values->next->next->vals[vrndx].a = (chrary *) mymalloc(sizeof(chrary), _initvar_);		/* VALGLB */
				for (i=0; i<MAXSUBARY; i++)
					cap->subary[i] = (chrsubary *) NULL;
			}
			if (!alarmcurr->vars[vrndx].a) {
				sap = alarmcurr->vars[vrndx].a = (spcary *) mymalloc(sizeof(spcary), _initvar_);
				for (i=0; i<MAXSUBARY; i++)
					sap->subary[i] = (spcsubary *) NULL;
			}
			break;
		case NIL:
			break;
	}
}

/* register a scancolumn var
 */
void
regscvar (int vrndx, int arndx)
{
	ky *nkp;

	nkp = (ky *) mymalloc(sizeof(ky), "regscvar");
	nkp->next = (ky *) NULL;
	nkp->vrndx = vrndx;
	nkp->arndx = arndx;
	MAXARNDX(vrndx) = maxval(MAXARNDX(vrndx), arndx);
	if (!SCVARS)
		SCVARS = nkp;
	else
		scvarcurr->next = nkp;
	scvarcurr = nkp;
	return;
}

/* register a history lookup key
 */
void
regkey (char *key, int valtype, int vartype, int arndx)
{
	ky *nkp;

	nkp = (ky *) mymalloc(sizeof(ky), "regkey");
	nkp->next = (ky *) NULL;
	switch ((nkp->vrndx = varndx(key+1))) {
		case NIL:
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regkey(), %s, unknown key %s%s", ALMNAM, key, vartype ? "[]" : "");
			STATUS = INACTIVE;
			return;
			break;
		case ERROR:
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regkey(), %s, attempted use of reserved fn name %s as key", ALMNAM, key);
			STATUS = INACTIVE;
			return;
			break;
		case INLINNOW:
			break;
		/* varndx() never returns INLINPRC
		case INLINPRC:
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regkey(), %s, attempted use of @inlin as key", ALMNAM);
			STATUS = INACTIVE;
			return;
			break;
		 */
		case RDLINNOW:
		/* varndx() never returns RDLINPRC
		case RDLINPRC:
		 */
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regkey(), %s, attempted use of reserved var name %s as key", ALMNAM, key);
			STATUS = INACTIVE;
			return;
			break;
		default:
			if (ISARRAY(vartype)) {
				if (TRIM(nkp->vrndx,arndx) == NIL) {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regkey(), %s, unknown key %s[%d]", ALMNAM, key, arndx);
					STATUS = INACTIVE;
					return;
				}
			}
			break;
	}
	dohistory = TRUE;		/* so, process history file */
	HSTTYP |= valtype;
	switch (vartype) {
		case SCALAR:
			nkp->arndx = 0;
			break;
		case ARRAY:
			nkp->arndx = arndx;
			MAXARNDX(nkp->vrndx) = maxval(MAXARNDX(nkp->vrndx), arndx);
			break;
		case NIL:
			break;
		default:
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regkey(), %s, bad vartype (%d) for key %s", ALMNAM, vartype, key);
			STATUS = INACTIVE;
			return;
			break;
	}
	NEWHST(nkp->vrndx) = TRUE;
	if (!KEYS)
		KEYS = nkp;
	else
		keycurr->next = nkp;
	keycurr = nkp;
	return;
}

/* register a rule
 */
void
regrul (int type)
{
	rul *nrp;

	nrp = (rul *) mymalloc(sizeof(rul), "regrul");
	nrp->next = (rul *) NULL;
	nrp->statements = (stm *) NULL;
	switch (type) {
		case BEG:
			alarmcurr->begin = nrp;
			break;
		case RUL:
			if (!alarmcurr->rules)
				alarmcurr->rules = nrp;
			else
				rulcurr->next = nrp;
			break;
		case END:
			alarmcurr->end = nrp;
			break;
	}
	rulcurr = nrp;
	return;
}

/* register a token; if a var, replace its type char ($#?%@) with a signed
   char value indicating its type; replace the char immediately following
   the type char with an index from 1 to 255; so, for example, $foo might
   become the char sequence

      -1            4            o            o
       ^            ^            ^            ^
     VLNOW        4th var    unchanged    unchanged

   add the token to either the lhsexpr or rshexpr token list for the
   current statement
 */
void
regtok (signed char *token, int isfn)
{
	static char _regtok_[] = "regtok";
	tok *ntp;
	signed int vrndx;
	int ndx;

	vrndx = NIL;
	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "regtok'ing %s", token);
	if (!isfn) {
		if ((*token == '$') || (*token == '#') || (*token == '?')) {
			if (isalnum((int)*(token+1))) {
				vrndx = varndx(token+1);
				switch (vrndx) {
					case VLPRS:
						ndx = atoi(token+1);
						if ((ndx < 0) || (ndx > MAXARYNDX)) {
							logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regtok(), %s, out-of-bounds value: %s", ALMNAM, token);
							STATUS = INACTIVE;
							return;
						} else {
							*token = VLPRS;
							*(token+1) = (unsigned char) ndx;
						}
						break;
					case INLINNOW:
						*token = INLINNOW;
						break;
					case RDLINNOW:
						*token = RDLINNOW;
						break;
					case ERROR:
					case NIL:
						/* error reported below */
						break;
					default:
						*token = VLNOW;
						*(token+1) = (unsigned char) vrndx;
						break;
				}
			}
		} else if ((*token == '%') && isalpha((int)*(token+1))) {
			vrndx = varndx(token+1);
			if (vrndx == NIL) {
				vrndx = freevarslot();
				if (vrndx == NIL) {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regtok(), %s, maximum var count (%d) exceeded for var %%%s%s", ALMNAM, MAXVARS, token, *(token+2) == '\[' ? "[]" : "");
					STATUS = INACTIVE;
					return;
				} else
					VARNAM(vrndx) = mystrdup(token+1, _regtok_);
			}
			switch (vrndx) {
				case VLPRS:
				case INLINNOW:
				case RDLINNOW:
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regtok(), %s, improper history reference for var %%%s", ALMNAM, token);
					STATUS = INACTIVE;
					return;
					break;
				case ERROR:
				case NIL:
					/* error reported below */
					break;
				default:
					*token = VLHST;
					*(token+1) = (unsigned char) vrndx;
					reghst(vrndx);
					break;
			}
		} else if ((*token == '@') && isalpha((int)*(token+1))) {
			vrndx = varndx(token+1);
			if (vrndx == NIL) {
				vrndx = freevarslot();
				if (vrndx == NIL) {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regtok(), %s, maximum var count (%d) exceeded for var @%s%s", ALMNAM, MAXVARS, token, *(token+2) == '\[' ? "[]" : "");
					STATUS = INACTIVE;
					return;
				} else
					VARNAM(vrndx) = mystrdup(token+1, _regtok_);
			}
			switch (vrndx) {
				case VLPRS:
				case ERROR:
				case NIL:
					/* error reported below */
					break;
				case INLINNOW:
					*token = INLINPRC;
					break;
				case RDLINNOW:
					*token = RDLINPRC;
					break;
				default:
					*token = VLPRC;
					*(token+1) = (unsigned char) vrndx;
					break;
			}
		} else if ((*token == '\[') && tokcurr && !strcmp(tokcurr->token,"$")) {
			/* ah, so this is an instance of $[...]; set the
			   first char of the previous token (still the
			   tokcurr before we reset it below) to VLPRS
			 */
			*tokcurr->token = VLPRS;
		} else {	/* token doesn't begin with $, #, %, @, or ? */
			vrndx = NIL;
		}
		if ((*token < 0) && ((signed char) *token != SQTPROXY) /* && ((signed char) *token != DQTPROXY) */ ) {
			if (vrndx == NIL) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regtok(), %s, unknown var %s", ALMNAM, token+1);
				STATUS = INACTIVE;
				return;
			}
			if (vrndx == ERROR) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regtok(), %s, attempted use of reserved fn name %s as var", ALMNAM, token+1);
				STATUS = INACTIVE;
				return;
			}
		}
	}
	ntp = (tok *) mymalloc(sizeof(tok), _regtok_);
	ntp->next = (tok *) NULL;
	ntp->token = token;
	switch (exprtyp) {
		case LHS:
			if (!stmcurr->lhsexpr)
				stmcurr->lhsexpr = ntp;
			else
				tokcurr->next = ntp;
			break;
		case RHS:
			if (!stmcurr->rhsexpr)
				stmcurr->rhsexpr = ntp;
			else
				tokcurr->next = ntp;
			break;
		default:
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in regtok(), %s, bad exprtyp for token %s", ALMNAM, token);
			STATUS = INACTIVE;
			break;
	}
	tokcurr = ntp;
	return;
}

/* a switch is converted, in effect, into a series of if statements
   internally; for example

      switch #foo
         case 1
            ...
         case 2
            ...
         case 3
            ...
      endsw

   would be converted into

      if #foo == 1
         ...
      endif
      if #foo == 2
         ...
      endif
      if #foo == 3
         ...
      endif

   duptok() duplicates the token sequence "#foo ==" (from the switch
   statement) for each of the newly created if statements
 */
void
duptok (tok *source, tok **target)
{
	tok *tp, *ntp;

	foreach (tp, source) {
		ntp = (tok *) mymalloc(sizeof(tok), "duptok");
		ntp->next = (tok *) NULL;
		ntp->token = tp->token;
		if (!*target)
			*target = ntp;
		else
			tokcurr->next = ntp;
		tokcurr = ntp;
	}
}

/* register a program statement: initialize data structures; for certain
   statements, create an auxiliary dummy statement (for holding essential
   control structure info); link it to the statement list
 */
void
regstm (int type)
{
	static char _regstm_[] = "regstm";
	stm *nsp, *asp;

	nsp = (stm *) mymalloc(sizeof(stm), _regstm_);
	nsp->tnext = (stm *) NULL;
	nsp->fnext = (stm *) NULL;
	nsp->type = type;
	nsp->status = ACTIVE;
	nsp->rhsexpr = (tok *) NULL;
	nsp->vrndx = 0;
	nsp->lhsexpr = (tok *) NULL;
	nsp->arndx = 0;
	nsp->logfile = (char *) NULL;
	if ((type == IF) || (type == WHILE) || (type == REPEAT) || (type == FOR) || (type == SWITCH)) {
		asp = (stm *) mymalloc(sizeof(stm), _regstm_);
		asp->tnext = (stm *) NULL;
		asp->fnext = (stm *) NULL;
		switch (type) {
			case IF:
				asp->type = ENDIF;
				break;
			case WHILE:
				asp->type = WHILEAUX;
				break;
			case REPEAT:
				asp->type = REPEATAUX;
				break;
			case FOR:
				asp->type = FORAUX;
				break;
			case SWITCH:
				asp->type = ENDSWITCH;
				break;
		}
		asp->rhsexpr = (tok *) NULL;
		asp->vrndx = 0;
		asp->lhsexpr = (tok *) NULL;
		asp->arndx = 0;
		asp->logfile = (char *) NULL;
		asp->aux = (stm *) NULL;
		nsp->aux = asp;
	} else
		nsp->aux = (stm *) NULL;
	if (!rulcurr->statements)
		rulcurr->statements = nsp;
	else
		stmcurr->next = stmcurr->tnext = nsp;
	if (stmcurr && (stmcurr->type == CASE)) {	/* previous statement is CASE */
		if ((type == CASE) || (type == DEFAULT))	/* current statement is CASE or DEFAULT */
			stmcurr->fnext = nsp;	/* set prev statement fnext to current statement */
	}
	stmcurr = nsp;
	stmcurr->fnext = nsp->aux;
	return;
}

/* register some history-related info for the current history var
 */
void
reghst (int ndx)
{

	dohistory = TRUE;		/* so, process history file */
	HSTTYP |= VALTYP(ndx);
	NEWHST(ndx) = TRUE;
	return;
}

/* given a var name, returns its index (1 to 255, or one of the special
   indices)
 */
signed int
varndx (char *name)
{
	char n[MAXIDENT+2+1];
	register signed int i;

	if (!isalnum((int)name[0]))
		return NIL;
	if (isdigit((int)name[0]))
		return VLPRS;
	sprintf(n, " %s ", name);
	if (strstr(fnnumnames, n) || strstr(fnstrnames, n))
		return ERROR;	/* name reserved as function name */
	if (strstr(" inlin inline inputline ", n))
		return INLINNOW;	/* applies also to @inlin, etc. */
	if (strstr(" rdlin rdline readline ", n))
		return RDLINNOW;
	for (i=1; i<=MAXVARS; i++) {
		if (VARNAM(i)) {
			if (!strcmp(VARNAM(i), name))
				return i;
		}
	}
	return NIL;
}

/* return a var's dat type (STRING/NUMBER)
 */
int
vardattype (char *name)
{
	signed int i;

	if ((i = varndx(name+1)) == NIL) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in vardattype(), %s, unknown var %s", ALMNAM, name+1);
		STATUS = INACTIVE;
		return ERROR;
	} else if ((i == VLPRS) || (i == INLINNOW) || (i == RDLINNOW))
		return STRING;
	else if (i == ERROR)
		return ERROR;
	else if (DATTYP(i) == NIL) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in vardattype(), %s, unknown dattype for var %s", ALMNAM, name+1);
		STATUS = INACTIVE;
		return ERROR;
	} else
		return DATTYP(i);
}

/* return the next available index in the sequence 1 ... 255
 */
int
freevarslot (void)
{
	register int i;

	for (i=1; i<=MAXVARS; i++) {
		if (!VARNAM(i))
			return i;
	}
	return NIL;
}

/* create and initialize elements in a new char (val) subarray
 */
void
initchrsubary (chrsubary **subary)
{
	register int i;
	chrsubary *ncsap;

	ncsap = (chrsubary *) mymalloc(sizeof(chrsubary), "initchrsubary");
	for (i=0; i<MAXSUBNDX; i++)
		ncsap->val[i] = (char *) NULL;
	*subary = ncsap;
	return;
}


/* create and initialize elements in a new spec subarray
 */
void
initspcsubary (spcsubary **subary)
{
	register int i;
	spcsubary *nssap;

	nssap = (spcsubary *) mymalloc(sizeof(spcsubary), "initspcsubary");
	for (i=0; i<MAXSUBNDX; i++) {
		nssap->fldnum[i] = 0;
		nssap->begincol[i] = 0;
		nssap->endcol[i] = 0;
		nssap->trim[i] = NIL;
	}
	*subary = nssap;
	return;
}
