%{

/*
   alt.y -- PIKT alarm script parser

   $Header: /home/cvs/ucgsb/pikt/src/piktd/alt.y,v 1.11 1999/10/30 17:46:41 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA

   --------------------------------------------------------------------------

   the ifblk & elifs sections could probably be consolidated/abbreviated

 */

#include "../lib/common.h"
#include "piktd.h"

char _altparse_[] = "altparse";

#define DUMP(token)	if (isdebug) logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "@%s@", token)

int i;
int valtype, dattype;

int vdt[MAXVAP];	/* var dat type */

char n[MAXIDENT+2+1];

stm *sp;

int ifnum, whilenum, repeatnum, fornum, switchnum;
int loopnum;

stm *loopsps[MAXLOOPS+3];	/* +3, because MAXLOOPS #defined as
				   3*MAXIFS in common.h
				 */

int isif[MAXIFS+1];
stm *ifsps[MAXIFS+1], *elifsps[MAXIFS+1], *elsesps[MAXIFS+1];
stm *elifsasps[MAXIFS+1], *elifszsps[MAXIFS+1];
stm *pelifsps[MAXIFS+1];
stm *stmszsps[MAXIFS+1], *fstmszsps[MAXIFS+1];

stm *whilesps[MAXWHILES+1], *endwhilesps[MAXWHILES+1];

stm *repeatsps[MAXREPEATS+1], *untilsps[MAXREPEATS+1];

stm *forsps[MAXFORS+1], *endforsps[MAXFORS+1];
stm *continuesps[MAXFORS+1];

stm *switchsps[MAXSWITCHS+1], *casesps[MAXSWITCHS+1], *defaultsps[MAXSWITCHS+1];
stm *lcasesps[MAXSWITCHS+1], *caseblkzsps[MAXSWITCHS+1];

stm *stmzsp;
stm *ifblkzsp, *whileblkzsp, *repeatblkzsp, *forblkzsp, *switchblkzsp;

stm *breaksp, *continuesp;

stm *dosp, *againsp, *leavesp, *redosp, *nextsp, *skipsp, *lastsp, *pausesp, *quitsp, *diesp, *outputsp, *setsp, *execsp;

#define CHKIFNUM	if (ifnum > MAXIFS) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, MAXIFS (%d) exceeded", ALMNAM, MAXIFS); \
				STATUS = INACTIVE; \
			} \
			else if (ifnum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, ifnum < 0", ALMNAM); \
				STATUS = INACTIVE; \
			}

#define CHKWHILENUM	if (whilenum > MAXWHILES) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, MAXWHILES (%d) exceeded", ALMNAM, MAXWHILES); \
				STATUS = INACTIVE; \
			} \
			else if (whilenum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, whilenum < 0", ALMNAM); \
				STATUS = INACTIVE; \
			}

#define CHKREPEATNUM	if (repeatnum > MAXREPEATS) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, MAXREPEATS (%d) exceeded", ALMNAM, MAXREPEATS); \
				STATUS = INACTIVE; \
			} \
			else if (repeatnum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, repeatnum < 0", ALMNAM); \
				STATUS = INACTIVE; \
			}

#define CHKFORNUM	if (fornum > MAXFORS) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, MAXFORS (%d) exceeded", ALMNAM, MAXFORS); \
				STATUS = INACTIVE; \
			} \
			else if (fornum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, fornum < 0", ALMNAM); \
				STATUS = INACTIVE; \
			}

#define CHKSWITCHNUM	if (switchnum > MAXSWITCHS) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, MAXSWITCHS (%d) exceeded", ALMNAM, MAXSWITCHS); \
				STATUS = INACTIVE; \
			} \
			else if (switchnum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, switchnum < 0", ALMNAM); \
				STATUS = INACTIVE; \
			}

#define CHKLOOPNUM	if (loopnum > MAXLOOPS) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, MAXLOOPS (%d) exceeded", ALMNAM, MAXLOOPS); \
				STATUS = INACTIVE; \
			} \
			else if (loopnum < 0) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, loopnum < 0", ALMNAM); \
				STATUS = INACTIVE; \
			}

#define ENTERIF		ifnum++;  \
			if (ifnum > MAXIFS) { \
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, MAXIFS (%d) exceeded", ALMNAM, MAXIFS); \
				STATUS = INACTIVE; \
			}

#define EXITIF		isif[ifnum] = 0; \
			ifblkzsp = ifsps[ifnum]->aux; \
			ifsps[ifnum] = elifsps[ifnum] = elsesps[ifnum] = elifsasps[ifnum] = elifszsps[ifnum] = pelifsps[ifnum] = stmszsps[ifnum] = fstmszsps[ifnum] = (stm *) NULL; \
			ifnum--

#define ENTERWHILE	whilenum++; \
			loopnum++; \
			CHKWHILENUM;

#define EXITWHILE	whileblkzsp = whilesps[whilenum]->aux; \
			whilesps[whilenum] = endwhilesps[whilenum] = (stm *) NULL; \
			loopsps[loopnum] = (stm *) NULL; \
			whilenum--; \
			loopnum--

#define ENTERREPEAT	repeatnum++; \
			loopnum++; \
			CHKREPEATNUM;

#define EXITREPEAT	repeatblkzsp = repeatsps[repeatnum]->aux; \
			repeatsps[repeatnum] = untilsps[repeatnum] = (stm *) NULL; \
			loopsps[loopnum] = (stm *) NULL; \
			repeatnum--; \
			loopnum--


#define ENTERFOR	fornum++; \
			loopnum++; \
			CHKFORNUM;

#define EXITFOR		forblkzsp = forsps[fornum]->aux; \
			forsps[fornum] = endforsps[fornum] = continuesps[fornum] = (stm *) NULL; \
			loopsps[loopnum] = (stm *) NULL; \
			fornum--; \
			loopnum--

#define ENTERSWITCH	switchnum++; \
			CHKSWITCHNUM;

#define EXITSWITCH	switchblkzsp = switchsps[switchnum]->aux; \
			switchsps[switchnum] = casesps[switchnum] = defaultsps[switchnum] = lcasesps[switchnum] = caseblkzsps[switchnum] = (stm *) NULL; \
			switchnum--

/* set the queued cases tnext to endswitch */
#define LINK_QUEUED_CASES_TO_ENDSW \
			sp = switchsps[switchnum]->aux->aux; \
			while (sp != casesps[switchnum]) { \
				sp->tnext = switchsps[switchnum]->aux; \
				sp = sp->next; \
			} \
			sp->tnext = switchsps[switchnum]->aux

/* set the queued cases tnext to the first case statement */
#define LINK_QUEUED_CASES_TO_CURR_CASE \
			sp = switchsps[switchnum]->aux->aux; \
			while (sp != casesps[switchnum]) { \
				sp->tnext = casesps[switchnum]->tnext; \
				sp = sp->next; \
			}

/* set the queued cases tnext to the first default statement */
#define LINK_QUEUED_CASES_TO_CURR_DEFAULT \
			sp = switchsps[switchnum]->aux->aux; \
			while (sp != casesps[switchnum]) { \
				sp->tnext = defaultsps[switchnum]->tnext; \
				sp = sp->next; \
			} \
			sp->tnext = defaultsps[switchnum]->tnext

/* set the last-statement tnext in the previous caseblk to the first
   statement in the current caseblk
 */
#define LINK_PREV_CASE_TO_CURR_CASE \
			if (caseblkzsps[switchnum]) \
				caseblkzsps[switchnum]->tnext = lcasesps[switchnum]->tnext

/* set the last-statement tnext in the previous caseblk to the first
   statement in the current defaultblk
 */
#define LINK_PREV_CASE_TO_CURR_DEFAULT \
			if (caseblkzsps[switchnum]) \
				caseblkzsps[switchnum]->tnext = defaultsps[switchnum]->tnext

%}

%union {
	char *sval;
}

%token <sval> _STANZA
%token _STATUS _ACTIVE _INACTIVE
%token _LEVEL _EMERG _ALERT _CRIT _ERR _WARNING _NOTICE _INFO _DEBUG
%token _TASK
%token _INPUT _PROCESS _REGFILE _LOGFILE
%token _FILTER
%token _SEPARATORS _DAT _KEYS
%token _INIT
%token _BEGIN _END
%token _RULE
%token _DO
%token _IF _ELIF _ELSE _ENDIF
%token _WHILE _ENDWHILE _REPEAT _UNTIL _FOR _ENDFOR _BREAK _CONTINUE
%token _SWITCH _CASE _DEFAULT _BREAKSWITCH _ENDSWITCH
%token _SWITCHLP _FORLP
%token _AGAIN _LEAVE _REDO _NEXT _SKIP _LAST _PAUSE _QUIT _DIE
%token _OUTPUT _MAIL _PRINT _LOG _PIKTLOG _SYSLOG
%token _SET _EXEC _WAIT
%token <sval> _STR _STRVAR
%token <sval> _POSITION _RANGE
%token <sval> _VARNUM _VARSTR _VARHSTSTR _VARHSTNUM _VARPRCSTR _VARPRCNUM
%token <sval> _FLHNDL
%token <sval> _VALNUM _VALSTR
%token <sval> _NOP _SOP _FOP _NREL _SREL
%token <sval> _NUM _INT
%token <sval> _FNNUM _FNSTR
%token _NOT _AND _OR

%type <sval> position varnum varstr varnow

%left _OR
%left _AND
%nonassoc _NREL _SREL
%left '+' '-' _NOP _SOP
%nonassoc _FOP
%nonassoc _NOT '~' _UNOP

%expect 5

%%

start:		alarmdefs
		;

alarmdefs:	alarmdef
		| alarmdefs alarmdef
		;

alarmdef:	stanza begin {
			if (!prescan) {
				STATUS = ACTIVE;
				INPUTTYP = NOINPUT;
			}
		}
		| stanza end {
			if (!prescan) {
				STATUS = ACTIVE;
				INPUTTYP = NOINPUT;
			}
		}
		| stanza begin end {
			if (!prescan) {
				STATUS = ACTIVE;
				INPUTTYP = NOINPUT;
			}
		}
		| stanza init begin {
			if (!prescan) {
				INPUTTYP = NOINPUT;
			}
		}
		| stanza init end {
			if (!prescan) {
				INPUTTYP = NOINPUT;
			}
		}
		| stanza init begin end {
			if (!prescan) {
				INPUTTYP = NOINPUT;
			}
		}
		| stanza init dataspec begin
		| stanza init dataspec end
		| stanza init dataspec begin end
		| stanza init dataspec rules
		| stanza init dataspec begin rules
		| stanza init dataspec rules end
		| stanza init dataspec begin rules end
		;

stanza:		_STANZA {
			if (!prescan) {
				DUMP($1);
				if (!alarmcurr)
					alarmcurr = alarms;
				else
					alarmcurr = alarmcurr->next;
				free($1);
			} else {
				regalm($1);
			}
		}
		;

init:		_INIT {
			if (!prescan) {
				if (interactive)
					STATUS = ACTIVE;
				else {
					logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, missing status, level & task", ALMNAM);
				}
			}
		}
		| _INIT status level task
		;

status:		_STATUS _ACTIVE { if (!prescan)	STATUS = ACTIVE; }
		| _STATUS _INACTIVE { if (!prescan) STATUS = INACTIVE; }
		;
		
level:		_LEVEL _EMERG { if (!prescan) LEVEL = LOG_EMERG; }
		| _LEVEL _ALERT { if (!prescan) LEVEL = LOG_ALERT; }
		| _LEVEL _CRIT { if (!prescan) LEVEL = LOG_CRIT; }
		| _LEVEL _ERR { if (!prescan) LEVEL = LOG_ERR; }
		| _LEVEL _WARNING { if (!prescan) LEVEL = LOG_WARNING; }
		| _LEVEL _NOTICE { if (!prescan) LEVEL = LOG_NOTICE; }
		| _LEVEL _INFO { if (!prescan) LEVEL = LOG_INFO; }
		| _LEVEL _DEBUG { if (!prescan) LEVEL = LOG_DEBUG; }
		;

task:		_TASK _STR {
			if (!prescan) {
				dequote($2);
				DUMP($2);
				TASK = $2;
			} else
				free($2);
		}
		;

dataspec:	input filter separators inputdata keys
		;

input:		_INPUT _PROCESS _STR {
			if (!prescan) {
				DUMP($3);
				INPUTTYP = PROCINPUT;
				INPUTSRC = $3;
			} else
				free($3);
		}
		| _INPUT _REGFILE _STR {
			if (!prescan) {
				DUMP($3);
				INPUTTYP = FILEINPUT;
				INPUTSRC = $3;
			} else
				free($3);
		}
		| _INPUT _LOGFILE _STR {
			if (!prescan) {
				DUMP($3);
				INPUTTYP = LOGINPUT;
				INPUTSRC = $3;
				INPUTLOGNAME = $3;
				dohistory = TRUE;	/* so, process history file */
				HSTTYP |= LOGINPUT;
			} else
				free($3);
		}
		;

filter:		/* empty */ { ; }
		| _FILTER _STR {
			if (!prescan) {
				DUMP($2);
				FILTER = $2;
			} else
				free($2);
		}

separators:	/* empty */ {
			if (!prescan) {
				SEPARATORS = mystrdup(" \t", _altparse_);	/* space/tab */
			}
		}
		| _SEPARATORS _STR {
			if (!prescan) {
				DUMP($2);
				SEPARATORS = $2;
			} else
				free($2);
		}
		;

inputdata:	/* empty */
		| data
		;

data:		datum
		| data datum
		;

datum:		_DAT varnow position {
			valtype = FIXEDVAL;
			if (!prescan) {
				DUMP($2);
				DUMP($3);
			} else {
				regvar($2+1, valtype, SCALAR, dattype, 0, $3);
			}
			free($2);
			free($3);
		}
		| _DAT varnow '[' _INT ']' position {
			valtype = FIXEDVAL;
			if (!prescan) {
				DUMP($2);
				DUMP("[");
				DUMP($4);
				DUMP("]");
				DUMP($6);
			} else {
				regvar($2+1, valtype, ARRAY, dattype, atoi($4), $6);
			}
			free($2);
			free($4);
			free($6);
		}
		| _DAT _STR {
			if (!prescan) {
				POSTYP = REGEXPR;
				REGEXP = mystrdup(stripesc($2), _altparse_);
			}
			free($2);
		}
		;

varnow:		_VARNUM { dattype = NUMBER; $$ = $1; }
		| _VARSTR { dattype = STRING; $$ = $1; }
		;

position:	_POSITION {
			/* DUMP($1); dumped in datum above */
			$$ = $1;
		}
		| _RANGE {
			DUMP($1);
			$$ = $1;
		}
		;

keys:		/* empty */ { ;	}
		| _KEYS keyset
		;

keyset:		varnow {
			if (!prescan) {
				DUMP($1);
				regkey($1, FIXEDVAL|FREEVAL, SCALAR, 0);
			}
			free($1);
		}
		| varnow '[' _INT ']' {
			if (!prescan) {
				DUMP($1);
				DUMP("[");
				DUMP($3);
				DUMP("]");
				regkey($1, FIXEDVAL|FREEVAL, ARRAY, atoi($3));
			}
			free($1);
			free($3);
		}
		| keyset varnow {
			if (!prescan) {
				DUMP($2);
				regkey($2, FIXEDVAL|FREEVAL, SCALAR, 0);
			}
			free($2);
		}
		| keyset varnow '[' _INT ']' {
			if (!prescan) {
				DUMP($2);
				DUMP("[");
				DUMP($4);
				DUMP("]");
				regkey($2, FIXEDVAL|FREEVAL, ARRAY, atoi($4));
			}
			free($2);
			free($4);
		}
		;

begin:		_BEGIN {
			valtype = BEGINVAL;
			if (!prescan) {
				exprtyp = RHS;
				ifnum = whilenum = repeatnum = fornum = loopnum = switchnum = 0;
				regrul(BEG);
			}
		}
		statements {
			/*
			if (!prescan) {
				if (isdebug)
					dumpsps(alarmcurr->begin->statements, "leaving _BEGIN");
			}
			*/
			;
		}
		;

end:		_END {
			valtype = ENDVAL;
			if (!prescan) {
				exprtyp = RHS;
				ifnum = whilenum = repeatnum = fornum = loopnum = switchnum = 0;
				regrul(END);
			}
		}
		statements {
			/*
			if (!prescan) {
				if (isdebug)
					dumpsps(alarmcurr->end->statements, "leaving _END");
			}
			*/
			;
		}
		;

rules:		rule
		| rules rule
		;

rule:		_RULE {
			valtype = FREEVAL;
			if (!prescan) {
				exprtyp = RHS;
				ifnum = whilenum = repeatnum = fornum = loopnum = switchnum = 0;
				regrul(RUL);
			}
		}
		statements {
			/*
			if (!prescan) {
				if (isdebug)
					dumpsps(alarmcurr->rules->statements, "leaving _RULE");
			}
			*/
			;
		}
		;

statements:	statement {
			if (!prescan) {
				if (ifnum >= 0) {
					stmszsps[ifnum] = stmzsp;
					if (isif[ifnum])
						fstmszsps[ifnum] = stmzsp;
				}
			}
		}
		| statements statement {
			if (!prescan) {
				if (ifnum >= 0) {
					stmszsps[ifnum] = stmzsp;
					if (isif[ifnum])
						fstmszsps[ifnum] = stmzsp;
				}
			}
		}
		;

statement:	do { if (!prescan) stmzsp = dosp; }
		| ifblk { if (!prescan) stmzsp = ifblkzsp; }
		| whileblk { if (!prescan) stmzsp = whileblkzsp; }
		| repeatblk { if (!prescan) stmzsp = repeatblkzsp; }
		| forblk { if (!prescan) stmzsp = forblkzsp; }
		| switchblk { if (!prescan) stmzsp = switchblkzsp; }
		| break { if (!prescan) stmzsp = breaksp; }
		| continue { if (!prescan) stmzsp = continuesp; }
		| again { if (!prescan) stmzsp = againsp; }
		| leave { if (!prescan) stmzsp = leavesp; }
		| redo { if (!prescan) stmzsp = redosp; }
		| next { if (!prescan) stmzsp = nextsp; }
		| skip { if (!prescan) stmzsp = skipsp; }
		| last { if (!prescan) stmzsp = lastsp; }
		| pause { if (!prescan) stmzsp = pausesp; }
		| quit { if (!prescan) stmzsp = quitsp; }
		| die { if (!prescan) stmzsp = diesp; }
		| output { if (!prescan) stmzsp = outputsp; }
		| set { if (!prescan) stmzsp = setsp; }
		| exec { if (!prescan) stmzsp = execsp; }
		;

do:		_DO {
			if (!prescan) {
				regstm(DO);
				dosp = stmcurr;
			}
		} expstr
		| _DO {
			if (!prescan) {
				regstm(DO);
				dosp = stmcurr;
			}
		} expnum
		;

ifblk:		if endif {
			if (!prescan) {
				EXITIF;
			}
		}
		| if statements endif {
			if (!prescan) {
				EXITIF;
			}
		}
		| if else endif {
			if (!prescan) {
				ifsps[ifnum]->tnext = ifsps[ifnum]->aux;
				ifsps[ifnum]->fnext = elsesps[ifnum];
				EXITIF;
			}
		}
		| if statements else endif {
			if (!prescan) {
				ifsps[ifnum]->fnext = elsesps[ifnum];
				stmszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				EXITIF;
			}
		}
		| if else statements endif {
			if (!prescan) {
				ifsps[ifnum]->tnext = ifsps[ifnum]->aux;
				ifsps[ifnum]->fnext = elsesps[ifnum];
				EXITIF;
			}
		}
		| if statements else statements endif {
			if (!prescan) {
				ifsps[ifnum]->fnext = elsesps[ifnum];
				fstmszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				EXITIF;
			}
		}
		| if elifs endif {
			if (!prescan) {
				ifsps[ifnum]->tnext = ifsps[ifnum]->aux;
				ifsps[ifnum]->fnext = elifsasps[ifnum];
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				elifszsps[ifnum]->fnext = ifsps[ifnum]->aux;
				EXITIF;
			}
		}
		| if statements elifs endif {
			if (!prescan) {
				ifsps[ifnum]->fnext = elifsasps[ifnum];
				fstmszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				elifszsps[ifnum]->fnext = ifsps[ifnum]->aux;
				EXITIF;
			}
		}
		| if elifs else endif {
			if (!prescan) {
				ifsps[ifnum]->tnext = ifsps[ifnum]->aux;
				ifsps[ifnum]->fnext = elifsasps[ifnum];
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				pelifsps[ifnum]->fnext = elsesps[ifnum];
				EXITIF;
			}
		}
		| if statements elifs else endif {
			if (!prescan) {
				ifsps[ifnum]->fnext = elifsasps[ifnum];
				fstmszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				pelifsps[ifnum]->fnext = elsesps[ifnum];
				EXITIF;
			}
		}
		| if elifs else statements endif {
			if (!prescan) {
				ifsps[ifnum]->tnext = ifsps[ifnum]->aux;
				ifsps[ifnum]->fnext = elifsasps[ifnum];
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				pelifsps[ifnum]->fnext = elsesps[ifnum];
				EXITIF;
			}
		}
		| if statements elifs else statements endif {
			if (!prescan) {
				ifsps[ifnum]->fnext = elifsasps[ifnum];
				fstmszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				pelifsps[ifnum]->fnext = elsesps[ifnum];
				EXITIF;
			}
		}
		;

elifs:		elif {
			if (!prescan) {
				pelifsps[ifnum] = elifsps[ifnum];
				elifsasps[ifnum] = elifsps[ifnum];
				elifszsps[ifnum] = elifsps[ifnum];
			}
		}
		| elif statements {
			if (!prescan) {
				elifsps[ifnum]->fnext = ifsps[ifnum]->aux;
				pelifsps[ifnum] = elifsps[ifnum];
				elifsasps[ifnum] = elifsps[ifnum];
				elifszsps[ifnum] = stmszsps[ifnum];
			}
		}
		| elifs elif {
			if (!prescan) {
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				pelifsps[ifnum]->fnext = elifsps[ifnum];
				pelifsps[ifnum] = elifsps[ifnum];
				elifszsps[ifnum] = elifsps[ifnum];
			}
		}
		| elifs elif statements {
			if (!prescan) {
				elifszsps[ifnum]->tnext = ifsps[ifnum]->aux;
				pelifsps[ifnum]->fnext = elifsps[ifnum];
				pelifsps[ifnum] = elifsps[ifnum];
				elifsps[ifnum]->fnext = ifsps[ifnum]->aux;
				elifszsps[ifnum] = stmszsps[ifnum];
			}
		}
		;

if:		_IF {
			if (!prescan) {
				ENTERIF;
				regstm(IF);
				ifsps[ifnum] = stmcurr;
				isif[ifnum] = TRUE;
			}
		}
		conditions
		;

elif:		_ELIF {
			if (!prescan) {
				regstm(ELIF);
				elifsps[ifnum] = stmcurr;
				isif[ifnum] = FALSE;
			}
		}
		conditions
		;

else:		_ELSE {
			if (!prescan) {
				regstm(ELSE);
				elsesps[ifnum] = stmcurr;
				isif[ifnum] = FALSE;
			}
		}
		;

endif:		_ENDIF {
			if (!prescan) {
				stmcurr->next = stmcurr->tnext = ifsps[ifnum]->aux;
				stmcurr = ifsps[ifnum]->aux;
				isif[ifnum] = FALSE;
			}
		}
		;

whileblk:	while endwhile {
			if (!prescan) {
				endwhilesps[whilenum]->tnext = whilesps[whilenum];
				EXITWHILE;
			}
		}
		| while statements endwhile {
			if (!prescan) {
				endwhilesps[whilenum]->tnext = whilesps[whilenum];
				EXITWHILE;
			}
		}
		;

while:		_WHILE {
			if (!prescan) {
				ENTERWHILE;
				regstm(WHILE);
				whilesps[whilenum] = loopsps[loopnum] = stmcurr;
			}
		}
		conditions
		;

endwhile:	_ENDWHILE {
			if (!prescan) {
				regstm(ENDWHILE);
				endwhilesps[whilenum] = stmcurr;
				stmcurr->next = whilesps[whilenum]->aux;
				stmcurr = whilesps[whilenum]->aux;
			}
		}
		;

repeatblk:	repeat until {
			if (!prescan) {
				untilsps[repeatnum]->fnext = repeatsps[repeatnum];
				EXITREPEAT;
			}
		}
		| repeat statements until {
			if (!prescan) {
				untilsps[repeatnum]->fnext = repeatsps[repeatnum];
				EXITREPEAT;
			}
		}
		;

repeat:		_REPEAT {
			if (!prescan) {
				ENTERREPEAT;
				regstm(REPEAT);
				repeatsps[repeatnum] = loopsps[loopnum] = stmcurr;
			}
		}
		;

until:		_UNTIL {
			if (!prescan) {
				regstm(UNTIL);
				untilsps[repeatnum] = stmcurr;
			}
		}
		conditions {
			if (!prescan) {
				stmcurr->next = stmcurr->tnext = repeatsps[repeatnum]->aux;
				stmcurr = repeatsps[repeatnum]->aux;
			}
		}
		;

forblk:		for endfor {
			if (!prescan) {
				endforsps[fornum]->tnext = forsps[fornum];
				EXITFOR;
			}
		}
		| for statements endfor {
			if (!prescan) {
				endforsps[fornum]->tnext = forsps[fornum];
				EXITFOR;
			}
		}
		;

for:		_FOR _VARNUM '=' {
			if (!prescan) {
				ENTERFOR;
				DUMP($2);
				regstm(SET);
			}
			regvar($2+1, valtype, SCALAR, NUMBER, 0, (char *)NULL);
			free($2);
		} expnum {
			if (!prescan) {
				regstm(FOR);
				forsps[fornum] = loopsps[loopnum] = stmcurr;
			}
		} conditions
		_VARNUM '=' {
			if (!prescan) {
				stmcurr = forsps[fornum]->aux;
				DUMP($8);
			}
			regvar($8+1, valtype, SCALAR, NUMBER, 0, (char *)NULL);
			free($8);
		} expnum {
			if (!prescan) {
				stmcurr = forsps[fornum];
			}
		}
		| _FORLP _VARNUM '=' {
			if (!prescan) {
				ENTERFOR;
				DUMP($2);
				regstm(SET);
			}
			regvar($2+1, valtype, SCALAR, NUMBER, 0, (char *)NULL);
			free($2);
		} expnum {
			if (!prescan) {
				regstm(FOR);
				forsps[fornum] = loopsps[loopnum] = stmcurr;
			}
		} conditions
		_VARNUM '=' {
			if (!prescan) {
				stmcurr = forsps[fornum]->aux;
				DUMP($8);
			}
			regvar($8+1, valtype, SCALAR, NUMBER, 0, (char *)NULL);
			free($8);
		} expnum {
			if (!prescan) {
				stmcurr = forsps[fornum];
			}
		}
		')'
		;

endfor:		_ENDFOR {
			if (!prescan) {
				regstm(SET);
				stmcurr->rhsexpr = forsps[fornum]->aux->rhsexpr;
				stmcurr->vrndx = forsps[fornum]->aux->vrndx;
				forsps[fornum]->aux->rhsexpr = (tok *) NULL;
				if (continuesps[fornum]) {
					continuesps[fornum]->fnext = stmcurr;
					continuesps[fornum] = (stm *) NULL;
				}
				regstm(ENDFOR);
				endforsps[fornum] = stmcurr;
				stmcurr->next = forsps[fornum]->aux;
				stmcurr = forsps[fornum]->aux;
			}
		}
		;

switchblk:	switch cases endswitch {
			if (!prescan) {
				LINK_QUEUED_CASES_TO_ENDSW;
				EXITSWITCH;
			}
		}
		| switch caseblks cases endswitch {
			if (!prescan) {
				LINK_QUEUED_CASES_TO_ENDSW;
				EXITSWITCH;
			}
		}
		| switch caseblks endswitch {
			if (!prescan) {
				EXITSWITCH;
			}
		}
		| switch caseblks defaultblk endswitch {
			if (!prescan) {
				/* set fnext for previous case to default */
				if (lcasesps[switchnum])
					lcasesps[switchnum]->fnext = defaultsps[switchnum];
				EXITSWITCH;
			}
		}
		| switch defaultblk endswitch {
			if (!prescan) {
				EXITSWITCH;
			}
		}
		;

caseblks:	caseblk
		| caseblks caseblk
		;

caseblk:	cases statements {
			if (!prescan) {
				LINK_QUEUED_CASES_TO_CURR_CASE;
				LINK_PREV_CASE_TO_CURR_CASE;
				/* note last statement in this caseblk */
				caseblkzsps[switchnum] = stmzsp;
			}
		}
		| cases breakswitch {
			if (!prescan) {
				LINK_QUEUED_CASES_TO_ENDSW;
				caseblkzsps[switchnum] = (stm *) NULL;
			}
		}
		| cases statements breakswitch {
			if (!prescan) {
				LINK_QUEUED_CASES_TO_CURR_CASE;
				LINK_PREV_CASE_TO_CURR_CASE;
				caseblkzsps[switchnum] = (stm *) NULL;
			}
		}
		;

defaultblk:	default
		| default statements {
			if (!prescan) {
				LINK_PREV_CASE_TO_CURR_DEFAULT;
			}
		}
		| default breakswitch
		| default statements breakswitch {
			if (!prescan) {
				LINK_PREV_CASE_TO_CURR_DEFAULT;
			}
		}
		| cases default {
			if (!prescan) {
				LINK_QUEUED_CASES_TO_ENDSW;
			}
		}
		| cases default statements {
			if (!prescan) {
				LINK_PREV_CASE_TO_CURR_DEFAULT;
				LINK_QUEUED_CASES_TO_CURR_DEFAULT;
			}
		}
		| cases default breakswitch {
			if (!prescan) {
				LINK_QUEUED_CASES_TO_ENDSW;
			}
		}
		| cases default statements breakswitch {
			if (!prescan) {
				LINK_PREV_CASE_TO_CURR_DEFAULT;
				LINK_QUEUED_CASES_TO_CURR_DEFAULT;
			}
		}
		;

cases:		case {
			if (!prescan) {
				/* note first case in series of cases in endswitch aux */
				switchsps[switchnum]->aux->aux = casesps[switchnum];
				/* set fnext for previous case, if any, to this case */
				if (lcasesps[switchnum])
					lcasesps[switchnum]->fnext = casesps[switchnum];
				/* note latest case in series of cases */
				lcasesps[switchnum] = casesps[switchnum];
			}
		}
		| cases case {
			if (!prescan) {
				/* note latest case in series of cases */
				lcasesps[switchnum] = casesps[switchnum];
			}
		}
		;

switch:		_SWITCH {
			if (!prescan) {
				ENTERSWITCH;
				regstm(SWITCH);
				switchsps[switchnum] = stmcurr;
			}
		} expstr {
			if (!prescan) {
				regtok(mystrdup("eq", _altparse_), FALSE);
			}
		}
		| _SWITCHLP {
			if (!prescan) {
				ENTERSWITCH;
				regstm(SWITCH);
				switchsps[switchnum] = stmcurr;
			}
		} expstr ')' {
			if (!prescan) {
				regtok(mystrdup("eq", _altparse_), FALSE);
			}
		}
		| _SWITCH {
			if (!prescan) {
				ENTERSWITCH;
				regstm(SWITCH);
				switchsps[switchnum] = stmcurr;
			}
		} expnum {
			if (!prescan) {
				regtok(mystrdup("==", _altparse_), FALSE);
			}
		}
		| _SWITCHLP {
			if (!prescan) {
				ENTERSWITCH;
				regstm(SWITCH);
				switchsps[switchnum] = stmcurr;
			}
		} expnum ')' {
			if (!prescan) {
				regtok(mystrdup("==", _altparse_), FALSE);
			}
		}
		;

case:		_CASE {
			if (!prescan) {
				regstm(CASE);
				casesps[switchnum] = stmcurr;
				duptok(switchsps[switchnum]->rhsexpr, &stmcurr->rhsexpr);
				casesps[switchnum]->fnext = switchsps[switchnum]->aux;
			}
		} expstr
		| _CASE {
			if (!prescan) {
				regstm(CASE);
				casesps[switchnum] = stmcurr;
				duptok(switchsps[switchnum]->rhsexpr, &stmcurr->rhsexpr);
				casesps[switchnum]->fnext = switchsps[switchnum]->aux;
			}
		} expnum
		;

default:	_DEFAULT {
			if (!prescan) {
				regstm(DEFAULT);
				defaultsps[switchnum] = stmcurr;
			}
		}
		;

breakswitch:	_BREAKSWITCH {
			if (!prescan) {
				regstm(BREAKSWITCH);
				stmcurr->fnext = switchsps[switchnum]->aux;
			}
		}
		;

endswitch:	_ENDSWITCH {
			if (!prescan) {
				stmcurr->next = stmcurr->tnext = switchsps[switchnum]->aux;
				stmcurr = switchsps[switchnum]->aux;
			}
		}
		;

break:		_BREAK {
			if (!prescan) {
				if (loopnum == 0)
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, break with no corresponding loop", ALMNAM);
				else
					CHKLOOPNUM;
				regstm(BREAK);
				stmcurr->fnext = loopsps[loopnum]->aux;
				breaksp = stmcurr;
			}
		}
		;

continue:	_CONTINUE {
			if (!prescan) {
				if (loopnum == 0)
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, continue with no corresponding loop", ALMNAM);
				else
					CHKLOOPNUM;
				regstm(CONTINUE);
				if (loopsps[loopnum]->type == FOR) {	/* a for loop */
					if (!continuesps[fornum])	/* first continue */
						continuesps[fornum] = stmcurr;
					else	/* subsequent continue, link to first continue */
						stmcurr->fnext = continuesps[fornum];
				} else {	/* not a for loop */
					stmcurr->fnext = loopsps[loopnum];
				}
				continuesp = stmcurr;
			}
		}
		;

again:		_AGAIN {
			if (!prescan) {
				regstm(AGAIN);
				againsp = stmcurr;
			}
		}
		;

leave:		_LEAVE {
			if (!prescan) {
				regstm(LEAVE);
				leavesp = stmcurr;
			}
		}
		;

redo:		_REDO {
			if (!prescan) {
				regstm(REDO);
				redosp = stmcurr;
			}
		}
		;

next:		_NEXT {
			if (!prescan) {
				regstm(NEXT);
				nextsp = stmcurr;
			}
		}
		;

skip:		_SKIP {
			if (!prescan) {
				regstm(SKIP);
				skipsp = stmcurr;
			}
		}
		expnum
		;

last:		_LAST {
			if (!prescan) {
				regstm(LAST);
				lastsp = stmcurr;
			}
		}
		;

pause:		_PAUSE {
			if (!prescan) {
				regstm(PAUSE);
				pausesp = stmcurr;
			}
		} expnum
		;

quit:		_QUIT {
			if (!prescan) {
				regstm(QUIT);
				quitsp = stmcurr;
			}
		}
		;

die:		_DIE {
			if (!prescan) {
				regstm(DIE);
				quitsp = stmcurr;
			}
		} expstr
		;

output:		_OUTPUT {
			if (!prescan) {
				regstm(OUTPUT);
				outputsp = stmcurr;
			}
		}
		expstr
		| _OUTPUT _MAIL {
			if (!prescan) {
				regstm(MAIL);
				outputsp = stmcurr;
			}
		}
		expstr
		| _OUTPUT _PRINT {
			if (!prescan) {
				regstm(PRINT);
				outputsp = stmcurr;
			}
		} expstr
		| _OUTPUT _LOG {
			if (!prescan) {
				regstm(LOCLOG);
				outputsp = stmcurr;
			}
		}
		_STR {
			if (!prescan) {
				DUMP($4);
				stmcurr->logfile = $4;
			} else
				free($4);
		}
		expstr
		| _OUTPUT _PIKTLOG {
			if (!prescan) {
				regstm(PIKTLOG);
				outputsp = stmcurr;
			}
		}
		expstr
		| _OUTPUT _SYSLOG {
			if (!prescan) {
				regstm(SYSLOG);
				outputsp = stmcurr;
			}
		}
		expstr
		;

set:		_SET _VARNUM '=' {
			if (!prescan) {
				DUMP($2);
				regstm(SET);
				setsp = stmcurr;
			}
			regvar($2+1, valtype, SCALAR, NUMBER, 0, (char *)NULL);
			free($2);
		} conditions
		| _SET _VARSTR '=' {
			if (!prescan) {
				DUMP($2);
				regstm(SET);
				setsp = stmcurr;
			}
			regvar($2+1, valtype, SCALAR, STRING, 0, (char *)NULL);
			free($2);
		} expstr
		| _SET _VARNUM '[' {
			if (!prescan) {
				DUMP($2);
				DUMP("[");
				regstm(SET);
				setsp = stmcurr;
				exprtyp = LHS;
			}
			regvar($2+1, valtype, ARRAY, NUMBER, NIL, (char *)NULL);
			free($2);
		}
		expnum
		']' '=' {
			if (!prescan) {
				DUMP("]");
				exprtyp = RHS;
			}
		}
		conditions
		| _SET _VARSTR '[' {
			if (!prescan) {
				DUMP($2);
				DUMP("[");
				regstm(SET);
				setsp = stmcurr;
				exprtyp = LHS;
			}
			regvar($2+1, valtype, ARRAY, STRING, NIL, (char *)NULL);
			free($2);
		}
		expnum
		']' '=' {
			if (!prescan) {
				DUMP("]");
				exprtyp = RHS;
			}
		}
		expstr
		;

exec:		_EXEC {
			if (!prescan) {
				regstm(EXEC);
				execsp = stmcurr;
			}
		} expstr
		| _EXEC _WAIT {
			if (!prescan) {
				regstm(EXECWAIT);
				execsp = stmcurr;
			}
		} expstr
		;

varstr:		_VARSTR {
			/* DUMP($1) */
			$$ = $1;
		}
		| _VARHSTSTR {
			/* DUMP($1) */
			$$ = $1;
		}
		| _VARPRCSTR {
			/* DUMP($1) */
			$$ = $1;
		}
		;

varnum:		_VARNUM {
			/* DUMP($1) */
			$$ = $1;
		}
		| _VARHSTNUM {
			/* DUMP($1) */
			$$ = $1;
		}
		| _VARPRCNUM {
			/* DUMP($1) */
			$$ = $1;
		}
		;

args:		/* empty */ { ; }
		| arg
		| argcoms arg
		;

argcoms:	argcom
		| argcoms argcom
		;

argcom:		arg
		',' {
			if (!prescan) {
				regtok(mystrdup(",", _altparse_), FALSE);
			}
		}
		;

arg:		conditions
		| _FLHNDL {
			if (!prescan) {
				regtok($1, FALSE);
			}
			/* whether FILEHANDLE or PROCHANDLE doesn't yet
			   matter at this point, so just pass FILEHANDLE
			 */
			regvar($1+1, 0, NIL, FILEHANDLE, NIL, (char *)NULL);
			if (prescan) {
				free($1);
			}
		}
		;

fnnum:		_FNNUM {
			if (!prescan) {
				DUMP($1);
				regtok($1, TRUE);
			} else {
				if (ischeck || isdoublecheck) {
					sprintf(n, " %s ", $1+1);
				 	if (!strstr(fnnumnames, n)) {
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, %s is not a number function", ALMNAM, $1);
						STATUS = INACTIVE;
					}
				}
				free($1);
			}
		}
		'(' {
			if (!prescan) {
				regtok(mystrdup("(", _altparse_), FALSE);
			}
		}
		args
		')' {
			if (!prescan) {
				regtok(mystrdup(")", _altparse_), FALSE);
			}
		}
		;

fnstr:		_FNSTR {
			if (!prescan) {
				DUMP($1);
				regtok($1, TRUE);
			} else {
				if (ischeck || isdoublecheck) {
					sprintf(n, " %s ", $1+1);
					if (!strstr(fnstrnames, n) &&
					    !strstr(" inlin inline inputline ", n)) {
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in altparse(), %s, %s is not a string function", ALMNAM, $1);
						STATUS = INACTIVE;
					}
				}
				free($1);
			}
		}
		'(' {
			if (!prescan) {
				regtok(mystrdup("(", _altparse_), FALSE);
			}
		}
		args
		')' {
			if (!prescan) {
				regtok(mystrdup(")", _altparse_), FALSE);
			}
		}
		;

expstr:		_STR {
			if (!prescan) {
				DUMP($1);
				regtok(mystrdup("\"", _altparse_), FALSE);
				regtok($1, FALSE);
				regtok(mystrdup("\"", _altparse_), FALSE);
			} else
				free($1);
		}
		| _STRVAR {
			if (!prescan) {
				DUMP($1);
				regtok(mystrdup("\"", _altparse_), FALSE);
				regtok($1, TRUE);
				regtok(mystrdup("\"", _altparse_), FALSE);
			} else
				free($1);
		}
		| _VALSTR {
			if (!prescan) {
				DUMP($1);
				regtok(mystrdup("\"", _altparse_), FALSE);
				regtok($1, FALSE);
				regtok(mystrdup("\"", _altparse_), FALSE);
			} else
				free($1);
		}
		| _VALSTR '[' {
			if (!prescan) {
				DUMP($1);
				DUMP("[");
				switch ((vdt[vapndx] = STRING)) {
					case STRING:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok($1, FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok($1, FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok(mystrdup(err, _altparse_), FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
				}
				vapndx++;
			}
		}
		expnum
		']' {
			if (!prescan) {
				vapndx--;
				DUMP("]");
				switch (vdt[vapndx]) {
					case STRING:
						regtok(mystrdup("]", _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok(mystrdup("]", _altparse_), FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("]", _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
				}
			}
		}
		| varstr {
			if (!prescan) {
				DUMP($1);
				switch (vardattype($1)) {
					case STRING:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok($1, FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok($1, FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok(mystrdup(err, _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
				}
			}
		}
		| varstr '[' {
			if (!prescan) {
				DUMP($1);
				DUMP("[");
				switch ((vdt[vapndx] = vardattype($1))) {
					case STRING:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok($1, FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok($1, FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok(mystrdup(err, _altparse_), FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
				}
				vapndx++;
			}
		}	
		expnum
		']' {
			if (!prescan) {
				vapndx--;
				DUMP("]");
				switch (vdt[vapndx]) {
					case STRING:
						regtok(mystrdup("]", _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok(mystrdup("]", _altparse_), FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("]", _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
				}
			}
		}
		| fnstr
		| expstr
		_SOP {
			if (!prescan) {
				DUMP($2);
				regtok($2, FALSE);
			} else
				free($2);
		}
		expstr
		;

expnum:		_NUM {
			if (!prescan) {
				DUMP($1);
				regtok($1, FALSE);
			} else
				free($1);
		}
		| _INT {
			if (!prescan) {
				DUMP($1);
				regtok($1, FALSE);
			} else
				free($1);
		}
		| _VALNUM {
			if (!prescan) {
				DUMP($1);
				regtok($1, FALSE);
			} else
				free($1);
		}
		| varnum {
			if (!prescan) {
				DUMP($1);
				switch (vardattype($1)) {
					case STRING:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok($1, FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok($1, FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok(mystrdup(err, _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
				}
			}
		}
		| varnum '[' {
			if (!prescan) {
				DUMP($1);
				DUMP("[");
				switch ((vdt[vapndx] = vardattype($1))) {
					case STRING:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok($1, FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok($1, FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("\"", _altparse_), FALSE);
						regtok(mystrdup(err, _altparse_), FALSE);
						regtok(mystrdup("[", _altparse_), FALSE);
						break;
				}
				vapndx++;
			}
		}
		expnum
		']' {
			if (!prescan) {
				vapndx--;
				DUMP("]");
				switch (vdt[vapndx]) {
					case STRING:
						regtok(mystrdup("]", _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
					case NUMBER:
						regtok(mystrdup("]", _altparse_), FALSE);
						break;
					case ERROR:
					default:
						regtok(mystrdup("]", _altparse_), FALSE);
						regtok(mystrdup("\"", _altparse_), FALSE);
						break;
				}
			}
		}
		| fnnum
		| '(' {
			if (!prescan) {
				regtok(mystrdup("(", _altparse_), FALSE);
			}
		} expnum ')' {
			if (!prescan) {
				regtok(mystrdup(")", _altparse_), FALSE);
			}
		}
		| '+' {
			if (!prescan) {
				regtok(mystrdup("+", _altparse_), FALSE);
			}
		}
		expnum %prec _UNOP
		| '-' {
			if (!prescan) {
				regtok(mystrdup("-", _altparse_), FALSE);
			}
		}
		expnum %prec _UNOP
		| '~' {
			if (!prescan) {
				regtok(mystrdup("~", _altparse_), FALSE);
			}
		}
		expnum
		| expnum
		_NOP {
			if (!prescan) {
				DUMP($2);
				regtok($2, FALSE);
			} else
				free($2);
		}
		expnum
		| expnum
		'+' {
			if (!prescan) {
				regtok(mystrdup("+", _altparse_), FALSE);
			}
		}
		expnum
		| expnum
		'-' {
			if (!prescan) {
				regtok(mystrdup("-", _altparse_), FALSE);
			}
		}
		expnum
		;

conditions:	expression
		| relation
		| '[' {
			if (!prescan) {
				regtok(mystrdup("[", _altparse_), FALSE);
			}
		} conditions ')' {
			if (!prescan) {
				regtok(mystrdup(")", _altparse_), FALSE);
			}
		}
		| _NOT {
			if (!prescan) {
				regtok(mystrdup("!", _altparse_), FALSE);
			}
		} conditions
		| conditions _AND {
			if (!prescan) {
				regtok(mystrdup("&&", _altparse_), FALSE);
			}
		} conditions
		| conditions _OR {
			if (!prescan) {
				regtok(mystrdup("||", _altparse_), FALSE);
			}
		} conditions
		;

expression:	expnum
		| _FOP {
			if (!prescan) {
				DUMP($1);
				regtok($1, FALSE);
			} else
				free($1);
		} expstr
		| expstr
		;

relation:	expnum
		_NREL {
			if (!prescan) {
				DUMP($2);
				regtok($2, FALSE);
			} else
				free($2);
		}
		expnum
		| expstr
		_SREL {
			if (!prescan) {
				DUMP($2);
				regtok($2, FALSE);
			} else
				free($2);
		}
		expstr
		;

%%
